


#ifndef _pos_h_
#define _pos_h_

#include"itemset.h"
#include"seq.h"

#ifndef POS_SC
 #ifdef POS_SC_SHORT
  #define POS_SC unsigned short
  #define POS_SC_END 65536
  #define POS_SCF "%hd"
 #else
  #define POS_SC unsigned char
  #define POS_SC_END 256
  #define POS_SCF "%hu"
 #endif
#endif

#define POS_DEFAULT_SC_SKIP 1
#define POS_DEFAULT_HASH_BUFSIZ 10000000

typedef struct {
  int flag;   //flag
  SEQ *S;  //  pointer to sequence structure
  ITEMSET *I;

    // position list
  SEQ_LONG *pos_list;     // set of positions
  SEQ_LONG pos_list_end;  // array size od pos_list
  SEQ_LONG pos_list_num;  // current number of positions in position list
  SEQ_CHAR skip_chr;  // charactor to be ignored
  int skip_chr_ub;        // upper bound of #skip_chr in pattern string not to be ignored
  int skip_repeat_th, skip_repeat_check;

    // for sc's
  POS_SC *sc;     // score for each position
  int *str_sc;    // score for each string
  int sc_max;     // maximum value for sc
  int sc_th;      // threshold for sc
  int str_sc_max; // maximum value for str_sc
  int sc_skip;    // reduction sc; how many position shares sc
  int sc_ub;      // upper bound of to be output
  SEQ_ID str_sc_end;   // end of the str_sc array
  LONG *sc_cnt1, *sc_cnt2, *sc_cnt_all;
  int k_level;    // how many levels (perr values) one position (string) is evaluated
  int init_str_sc;   // initial value for str_sc
  POS_SC init_sc, sc_mark, sc_mark2; // initial value for sc

    // for bucket sort
  SEQ_LONG **let_cnt; // array for counting the appearance of each letter, for each depth
  QUEUE *jump;      // queue of indices for non-zero let_cnt, for each depth 
  int dep_max, dep_skip;      // maximum depth of bucket sort, and #max. letters to be processed at one bucket sort
  int *pos_dep;     // position (i-th letter) of each depth
  int hash_end;     // the maximum of the letters (hash keys) 
  int hash_buf_siz;     // the maximum buffer size for the hash tables
  SEQ_LONG *IDcnt, *IDcnt2;        // array for counting the pattern strings of each ID
  QUEUE IDjump;        // queue for substID's for non-zero IDcnt

    // partitiion of pattern string for 
  int pblock_num, pblock_num_org;    // #blocks in the paritition of pattern string 
  QUEUE ppartition;   // positions giving blocks, to be sorted
  int *ppartition_mark;  // mark for each partition
  
  int chunk, chunk_org;  // original of plen and perr, pattern division number
  int cur_chunk, chunk_head, chunk_tail;   // current operating number, its head and tail
  int *chunk_pos;  // position of chunks
  
  SEQ_LONG sep_pos;     // boundary for positions to be compared
  int partition_th;     // threshold value for partition_by
  QUEUE remain_block;   // the blocks have not been compared in multi-classification
} POS;

#define POS_REMAIN_BLOCK 8 // normalize positions of all strings
#define POS_SC_USE 16   // use sc
#define POS_STR_SC 32   // use str_sc (score for strings) instead of sc (score for positions)
#define POS_SC_COUNT 64   // count the number of strings for each position 
#define POS_STR_SC_COUNT 128   // count the number of strings for each position 
#define POS_SC_MIN 256     // compute smallest distance to others
#define POS_STR_SC_MIN 512     // compute smallest distance to others, for string
#define POS_SC_COUNT_INTERVAL 2048   // count the number of strings for each interval (increase count for all positions in an interval
#define POS_SC_OUTPUT_ONLY_MAX 16384  //  output positions of strings with maximum sc
#define POS_ONE_CHUNK_MATCH 32768 // output a pair if just one of the chunks matches
#define POS_SC_COUNT_INTERVAL_EXTEND 2097152 // count interval (#number of matches for each) with extending similar strings
// #define SEQ_STR_SC_COUNT 8192   // counting sc for each string

//crcdisk

QSORT_TYPE_HEADER(POS_SC, POS_SC)

/* initializatin, allocation, termination*/
void POS_init (POS *P);
void POS_alloc (POS *P, SEQ *S, ITEMSET *I);
void POS_end (POS *P);
void POS_multicore_dup (POS *P, POS *PP, int level);
void POS_multicore_end (POS *P, int level);

/* determine #subblocks */
void POS_mk_pos_dep (POS *P, int s);

/* compute the optimal partition according to maching prob. p */
int POS_opt_partition (SEQ *S, int flag);
void POS_mk_ppartition (POS *P, int num);
int POS_opt_superpartition (POS *P);


/* output pattern substrings beginning at each position in pos_list */
void POS_fprint_pos_list (FILE *fp, POS *P, SEQ_LONG *pos_list, SEQ_LONG t);

/* translate positions in the file to intervals, and write them to the file */
/* consecutive integers are considered to be an interval */
/* each line has to include at least one integer, and the output is the 
 pair of integers representing the beginning/ending position of an interval */
void POS_fprint_pos_substrings (POS *P, char *input_fname, char *output_fname);


/*  output sc's for all positions 
  if POS_SC_OUTPUT_ONLY_MAX, then output the position with sc[] = POS_SC_END-1 */
void SEQ_fprint_pos_sc (FILE *fp, POS *P);

/*  output sc's for all strings 
  if POS_SC_OUTPUT_ONLY_MAX, then output the position with sc[] = POS_SC_END-1 */
void SEQ_fprint_str_sc (FILE *fp, POS *P);

/* output a pair of positions to file */ 
void POS_fprint_positions (FILE2 *fp, SEQ *S, SEQ_LONG a, LONG b, int k, int flag);

/* output a pair of positions, with possibly strings */
void POS_fprint_pair (FILE2 *fp, POS *P, SEQ_LONG a, SEQ_LONG b, int k, int core_id);

/* read the values from sc file */
void POS_read_sc (POS *P, char *fname);

/* remove all positions from posision list which include skip_char's no less than skip_chr_ub
   and repeat strings with Hamming distance at most th */
/* remove all positions with small sc, or small str_sc */
/* flag&1: str_sc, o.w. sc,   flag&2: remoce one's larger than th */
void POS_rm_pos_with_ignore_chr (POS *S);
void POS_rm_pos_repeated_string (POS *S, int th);
void POS_rm_pos_with_small_sc (POS *S, int flag, int th);

/* compute the 'letter' obtained by combining l characters, beginning from kth
 letter of position pos*/
int POS_bucket_letter (SEQ *S, SEQ_LONG pos, int k, int l);
int POS_bucket_letter_matrix (SEQ *S, SEQ_LONG pos, int k, int l);

/*  radix sort by kth letter
 positions in pos, up to t-th elements will be sorted, by k-th to (k+l-1)th char.
 The dep-th bucket and jump will be used */
void POS_bucket_sort (SEQ *S, SEQ_LONG *pos, SEQ_LONG t, int k, int l, QUEUE *jump, SEQ_LONG *cnt, SEQ_LONG *cnt2);
void POS_bucket_sort_by_substID (SEQ *S, SEQ_LONG *pos, SEQ_LONG t);

/* partition the positions from 0 to t-1 by substrings, those of ID<th, and those of ID>=th */
SEQ_LONG POS_partition_by_substID (SEQ *S, SEQ_LONG *pos, SEQ_LONG t);


/* compute optimal interleave
 set interleave_len, pos_skip_a,b */
/* set position list to all positions */
/* set position list to optimal positions for string comparison */
void SEQ_compute_interleave (SEQ *S);
void POS_mk_pos_list (POS *S);

/* compute the mismatches in other chunks */
int POS_check_chunk (POS *P, SEQ_CHAR *px, SEQ_CHAR *py, int k);
#ifdef SEQ_PPERM_MASK
int POS_remain_mismatch (POS *P, SEQ_LONG x, SEQ_LONG y, int block, int block_end);
#else
int POS_remain_mismatch (POS *P, SEQ_LONG x, SEQ_LONG y, int block, int block_end);
int POS_remain_mismatch_ (POS *P, SEQ_LONG x, SEQ_LONG y, int block, int block_end);
#endif

#endif


